/*
*****************************************************************************
  *		XpressNetMaster.h - library for XpressNet protocoll
  *		Copyright (c) 08/2016 - 2017 Philipp Gahtow  All right reserved.
  *
  *		Version 2.1.1
*****************************************************************************
  * FUNKTIONS:
  *
  *		Build a XpressNet Master for driving and switch
  *		Support for Roco LokMaus2 and Multimaus
  *		Basic support for direct CV programming
  *		Check XOR before packet dekoding
  * 
  * 	-> Please contact Lenz Inc. for more details about XpressNet.
  *
  * not yet Supported:
  *	- Emergency stop a locomotive: 0x92 AddrH AddrL [XOR]
  *	- Emergency stop a locomotive: 0x91 loco_addr [XOR] (v2)
  *	- Service Mode: only direct Mode
  *	- DCC extended accessory command: 0x13 0x01 B+AddrH AddrL
  *	- DCC FAST CLOCK: 0x01 0xF2 0xF3
  *	- Command Tunnel: 0x3* ** ** [XOR]
  *	- BiDi messages: 0x7* ** ** [XOR]
  *	- Library Entry: 0xE9 0xF1 AddrH AddrL IDX SIZE [NAME][XOR]
  *	- Set function state: 0xE4 0x2* AddrH AddrL Group [XOR]
*****************************************************************************
*/

// include this library's description file
#include "XpressNetMaster.h"
#include <avr/interrupt.h>

XpressNetMasterClass *XpressNetMasterClass::active_object = 0;	//Static

// Constructor /////////////////////////////////////////////////////////////////
// Function that handles the creation and setup of instances

XpressNetMasterClass::XpressNetMasterClass()
{
	// initialize this instance's variables 
	XNetAdr = 0;	//Startaddresse des ersten XNet Device
	XNetSlaveMode = 0x00;	//Start in MASTER MODE
	XNetSlaveInit = true;	//if change to Slave Mode make initialize
	Railpower = csTrackVoltageOff;
	Fahrstufe = Loco128;
	
	XNetBufferSend = 0;	//start position to read data from the buffer
	XNetBufferSend_data = 0;	//position of byte that we are sending
	XNetBufferStore = 0; //start position to store data in buffer
	
	for (byte s = 0; s < 32; s++) { //clear busy slots
		SlotLokUse[s] = 0xFFFF;	//slot is inactiv
	}

	for (byte b = 0; b < XNetBufferSize; b++) {	//clear send buffer
		XNetBuffer[b].length = 0x00;
		for (byte d = 0; d < XNetBufferMaxData; d++) {
			XNetBuffer[b].data[d] = 0x00;
		}
	}
	
	XNetDataReady = false;	//keine Daten empfangen!
	XNet_state = XNet_get_callbyte;	//set the start state
	XNetclear();	//alte Nachricht lschen
}

//******************************************Serial*******************************************
void XpressNetMasterClass::setup(uint8_t FStufen, uint8_t  XControl)  //Initialisierung Serial
{
	Fahrstufe = FStufen;
	MAX485_CONTROL = XControl;
	// LISTEN_MODE 
	pinMode(MAX485_CONTROL, OUTPUT);
	digitalWrite (MAX485_CONTROL, LOW);	//RECEIVE_MODE - SLAVE

	//Set up on 62500 Baud
	cli();  //disable interrupts while initializing the USART
	#ifdef SERIAL_PORT_0
	 UBRRH = 0;
	 UBRRL = 0x0F;
	 UCSRA = 0;
	 UCSRB = (1<<RXEN) | (1<<TXEN) | (1<<RXCIE) | (1<<TXCIE) | (1<<UCSZ2);
	 UCSRC = (1<<UCSZ1) | (1<<UCSZ0);
	#else
	/*
	 UBRR0H = 0;
	 UBRR0L = 0x0F;
	 UCSR0A = 0;
	 UCSR0B = (1<<RXEN0) | (1<<TXEN0) | (1<<RXCIE0) | (1<<TXCIE0)| (1<<UCSZ02);
	 UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
	 #else
	*/
	 UBRR1H = 0;
	 UBRR1L = 0x0F;
	 UCSR1A = 0;
	 UCSR1B = (1<<RXEN1) | (1<<TXEN1) | (1<<RXCIE1) | (1<<TXCIE1) | (1<<UCSZ12);
	 UCSR1C = (1<<UCSZ11) | (1<<UCSZ10);
	 #endif
	 sei(); // Enable the Global Interrupt Enable flag so that interrupts can be processed 
	 /*
	 *  Enable reception (RXEN = 1).
	 *  Enable transmission (TXEN0 = 1). 
	 *	Enable Receive Interrupt (RXCIE = 1).
	 *  Enable Transmit Interrupt (TXCIE = 1).
	 *  Set 9-bit character mode (UCSZ00, UCSZ01, and UCSZ02 together control this, 
	 *  But UCSZ00, UCSZ01 are in Register UCSR0C).
	 */
	 active_object = this;		//hold Object to call it back in ISR
}

// Public Methods //////////////////////////////////////////////////////////////
// Functions available in Wiring sketches, this library, and other libraries

//*******************************************************************************************
//Daten ermitteln und Auswerten
void XpressNetMasterClass::update(void) 
{
	//XpressNet State Machine:
	switch (XNet_state) {
		case XNet_send_callbyte:
			//wait until data sending is ready.
			break;
		case XNet_wait_receive:	//wait for client answer, max 120 microsekunden
			if (micros() - XSendCount >= XNetTimeUntilNext) {	//Timeout?
				XNet_state = XNet_get_callbyte;
			}
			if (XNetMsg[XNetlength] != 0x00) {
				#if defined (XNetDEBUGTime)
				XNetSerial.print("Reply after: ");
				XNetSerial.println(micros() - XSendCount);
				#endif
				XNet_state = XNet_receive_data;
				XSendCount = micros();	//save time for receive the first byte
			}
			break;
		case XNet_receive_data:	//read client data, max 500ms
			if (micros() - XSendCount >= XNetTimeReadData) {	//Timeout?
				XNet_state = XNet_get_callbyte;
				XNetclear();	//alte Nachricht lschen
				break;
			}
			if (!XNetDataReady)
				break;
			#if defined (XNetDEBUGTime)
			XNetSerial.print("Paket time: ");
			XNetSerial.println(micros() - XSendCount);
			#endif
			XNetDataReady = false;
			if (XNetCheckXOR())	//Checks the XOR
				XNetAnalyseReceived();	//Auswerten der empfangenen Daten
			XNetclear();	//alte Nachricht lschen
			if (XNetSlaveMode == 0x00) {		//MASTER MODE
				XNet_state = XNet_send_data;
				XNetSendNext();	//start sending out by interrupt
			}
			else XNet_state = XNet_get_callbyte;
			break;
		case XNet_send_data:
			break;
		case XNet_get_callbyte:	//address the next client
			if (XNetSlaveMode == 0x00) {		//MASTER MODE
				//we are not sending out more Data, get next CallByte
				XNet_state = XNet_send_callbyte;
				getNextXNetAdr();	//Send next CallByte
				XNetSendNext();	//start sending out by interrupt
			}
			else {
				//count cycles that we run in SLAVE MODE
				XNetSlaveMode--;	//stay only in SLAVE MODE if we receive CallBytes
				XNet_state = XNet_receive_data;
				//we need to initialize the SLAVE MODE?
				if (XNetSlaveInit) {
					XNetSlaveInit = false;
					uint8_t Init[] = { 0x00, 0x21, 0x24, 0x05 };	
					XNetsend(Init, 4);	//send initsequence
				}
				if (XNetSlaveMode == 0x00)	//we will leave the Slave-Mode?
					XNetSlaveInit = true;
			}
			
			XSendCount = micros(); //save time for callbyte send!
			break;
	}
}

//--------------------------------------------------------------------------------------------
//Checks the XOR
bool XpressNetMasterClass::XNetCheckXOR(void) {
	
	if (rxXOR == 0x00) {	//XOR is 0x00?
		return true;
	}
	//bertragungsfehler:
	if (XNetSlaveMode == 0x00) {		//MASTER MODE
		uint8_t ERR[] = {DirectedOps, 0x61, 0x80, 0xE1 };	
		XNetsend(ERR, 4);
	}
	return false;
}

//--------------------------------------------------------------------------------------------
//Daten Auswerten
void XpressNetMasterClass::XNetAnalyseReceived(void) {		//work on received data

			#if defined (XNetDEBUG)
			XNetSerial.print("XRX: 0x1");
			if (XNetSlaveMode == 0x00)
				XNetSerial.print(CallByteInquiry, HEX);	//MASTER MODE - CallByte 8 Bit only
			else XNetSerial.print(XNetMsgCallByte, HEX);	//SLAVE MODE - CallByte 8 Bit only
			for (byte i = 0; i < XNetMsg[XNetlength]; i++) {
				if (XNetMsg[i+1] < 0x10)
					XNetSerial.print(" 0x0");
				else XNetSerial.print(" 0x");
				XNetSerial.print(XNetMsg[i+1], HEX);
			}
			XNetSerial.println();
			#endif
			
			switch (XNetMsg[XNetheader]) {	
			case 0x21:
				if (XNetMsg[XNetdata1] == 0x24 && XNetMsg[XNetdata2] == 0x05) {		//Command station status indication response
					/*
					Bit 0: =1 - Command station is in emergency off (Nothalt)
					Bit 1: =1 - Command station is in emergency stop (Notaus)
					Bit 2: Command station-Start mode (0 = manual mode, 1 = automatic mode)
					Automatic Mode: All locomotives start at their last known speed each
					time the command station is powered up
					Manual Mode: All locomotives have a speed of 0 and functions out on
					command station power up
					Bit 3: = 1 - The command station is in service mode
					Bit 4: reserved
					Bit 5: reserved
					Bit 6: = 1 - The command station is performing a power up.
					Bit 7: = 1 - There was a RAM check error in the command station
					*/
					byte status = 0x01;	//csTrackVoltageOff = B1;
					switch (Railpower) {
						case csNormal:			status = 0; break;
						case csEmergencyStop:	status = 0x01; break;
						case csServiceMode:		status = 0x08; break;
						case csShortCircuit:	status = 0x02; break;
					}
					uint8_t sendStatus[] = { DirectedOps, 0x62, 0x22, status, 0x00 };
					getXOR(sendStatus, 5);
					XNetsend(sendStatus, 5);
				}
				if (XNetMsg[XNetdata1] == 0x21 && XNetMsg[XNetdata2] == 0x00) {		//Command station softwareversion response
					uint8_t sendVersion[] = { DirectedOps, 0x63, 0x21, XNetVersion, XNetID, 0x00 }; //63-21 36 0 74
					getXOR(sendVersion, 6);
					XNetsend(sendVersion, 6);
				}
				if (XNetMsg[XNetdata1] == 0x80 && XNetMsg[XNetdata2] == 0xA1) {		//Alles Aus (Notaus)
					setPower(csTrackVoltageOff);
				}
				if (XNetMsg[XNetdata1] == 0x81 && XNetMsg[XNetdata2] == 0xA0) {		//Alles An
					setPower(csNormal);
				}
				if (XNetMsg[XNetdata1] == 0x10 && XNetMsg[XNetdata2] == 0x31) {	//Request for Service Mode results 
					// Programming info. "Command station ready " 
					uint8_t sendStatus[] = { DirectedOps, 0x61, 0x11, 0x00 };
					if (Railpower != csServiceMode) {
						// Programming info. "Command station busy"
						sendStatus[2] = 0x1f;
					}
					getXOR(sendStatus, 4);
					XNetsend(sendStatus, 4);
				}
				if (SlotLokUse[DirectedOps & 0x1F] == 0xFFFF)
					SlotLokUse[DirectedOps & 0x1F] = 0;	//mark Slot as activ
				//XNetclear();	//alte Nachricht lschen
				break;
			case 0x22: //Start Programming
				if (XNetMsg[XNetdata1] == 0x11) {	//Register Mode read request (Register Mode) 
				}
				if (XNetMsg[XNetdata1] == 0x14) {	//Paged Mode read request (Paged Mode) 
				}
				if (XNetMsg[XNetdata1] == 0x15) {	//Direct Mode CV read request (CV mode) 
				}
				setPower(csServiceMode);
				//XNetclear();	//alte Nachricht lschen	
				break;
			case 0x23:
				if (XNetMsg[XNetdata1] == 0x12) { //Register Mode write request (Register Mode) 
				}
				if (XNetMsg[XNetdata1] == 0x16) { //Direct Mode CV write request (CV mode) 
					if (notifyXNetDirectCV)
						notifyXNetDirectCV(XNetMsg[XNetdata2], XNetMsg[XNetdata3]);
					break;
				}
				if (XNetMsg[XNetdata1] == 0x17) { //0x23 0x17 CV DAT[XOR] - Paged Mode write request(Paged mode)
					
				}
				unknown(); //unbekannte Anfrage
				//XNetclear();
				break; 
			case 0x80:
				if (XNetMsg[XNetdata1] == 0x80) {		//EmStop
					setPower(csEmergencyStop);
					break;
				}
				unknown(); //unbekannte Anfrage
				//XNetclear();	//alte Nachricht lschen
				break;
			case 0xE3: {
				switch (XNetMsg[XNetdata1]) {
					case 0x00: //Lokdaten anfordern & F0 bis F12 anfordern
								if (notifyXNetgiveLocoInfo)
									notifyXNetgiveLocoInfo(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));
								break;
					case 0x07: { //Funktionsstatus F0 bis F12 anfordern (Funktion ist tastend oder nicht tastend)
								//0x07, sonst ist LokMaus2 langsam!
								uint8_t LocoFkt[] = { DirectedOps, 0xE3, 0x50, 0x00, 0x00, 0x00 };
								getXOR(LocoFkt, 6);
								XNetsend(LocoFkt, 6);	
								break;
								}
					case 0x08: { //Funktionsstatus F13 bis F28 anfordern (Funktion ist tastend oder nicht tastend)
								uint8_t LocoFkt[] = { DirectedOps, 0xE3, 0x51, 0x00, 0x00, 0x00 };
								getXOR(LocoFkt, 6);
								XNetsend(LocoFkt, 6);
								break;
								}
					case 0x09: //Funktionszustand F13 bis F28 anfordern
								if (notifyXNetgiveLocoFunc)
									notifyXNetgiveLocoFunc(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));
								break;
					case 0xF0: //Lok und Funktionszustand MultiMaus anfordern
								if (notifyXNetgiveLocoMM)
									notifyXNetgiveLocoMM(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));
								break;
					default: unknown(); //unbekannte Anfrage
				}
				//XNetclear();	//alte Nachricht lschen
				break;
				}
			case 0xE4: {	//Fahrbefehle
				AddBusySlot(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));	//set Busy
				if (XNetMsg[XNetdata1] == 0x10) {	//14 Fahrstufen
					if (notifyXNetLocoDrive14)
						notifyXNetLocoDrive14(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x11) {	//27 Fahrstufen
					if (notifyXNetLocoDrive27)
						notifyXNetLocoDrive27(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x12) {	//28 Fahrstufen
					if (notifyXNetLocoDrive28)
						notifyXNetLocoDrive28(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x13) {	//128 Fahrstufen
					if (notifyXNetLocoDrive128)
						notifyXNetLocoDrive128(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x20) {	//Funktionsbefehl Gruppe1 0 0 0 F0 F4 F3 F2 F1
					if (notifyXNetLocoFunc1)
						notifyXNetLocoFunc1(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x21) {	//Funktionsbefehl Gruppe2 0000 F8 F7 F6 F5
					if (notifyXNetLocoFunc2)
						notifyXNetLocoFunc2(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x22) {	//Funktionsbefehl Gruppe3 0000 F12 F11 F10 F9
					if (notifyXNetLocoFunc3)
						notifyXNetLocoFunc3(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x23 || XNetMsg[XNetdata1] == 0xF3) {	//Funktionsbefehl Gruppe4 F20-F13
					//0xF3 = undocumented command is used when a mulitMAUS is controlling functions f20..f13. 
					if (notifyXNetLocoFunc4)
						notifyXNetLocoFunc4(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x28) {	//Funktionsbefehl Gruppe5 F28-F21
					if (notifyXNetLocoFunc5)
						notifyXNetLocoFunc5(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else unknown(); //unbekannte Anfrage
				break;
				}
			case 0x42:	//Accessory Decoder information request
				if (notifyXNetTrntInfo)
					notifyXNetTrntInfo(DirectedOps, XNetMsg[XNetdata1], XNetMsg[XNetdata2]);
				//XNetclear();	//alte Nachricht lschen
				break;
			case 0x52:	//Accessory Decoder operation request
				if (notifyXNetTrnt)
					notifyXNetTrnt((XNetMsg[XNetdata1] << 2) | ((XNetMsg[XNetdata2] & B110) >> 1), XNetMsg[XNetdata2]);
					//XNetMsg[XNetdata2] = 0000 ABBP
					//A = Weichenausgang(Spulenspannung EIN/AUS)
					//BB = Adresse des Dekoderport 1..4
					//P = Ausgang (Gerade = 0 / Abzweigen = 1)
				//XNetclear();	//alte Nachricht lschen
				break;
			default:  //Befehl in Zentrale nicht vorhanden
				unknown(); //unbekannte Anfrage
			}
	
	if (XNetSlaveMode != 0x00) {		//SLAVE-MODE
		if (XNetMsgCallByte == GENERAL_BROADCAST) {
			if (XNetMsg[XNetheader] == 0x61) {
			  if ((XNetMsg[XNetdata1] == 0x01) && (XNetMsg[XNetdata2] == 0x60)) {
				// Normal Operation Resumed
				setPower(csNormal);
			  } 
			  else if ((XNetMsg[XNetdata1] == 0x00) && (XNetMsg[XNetdata2] == 0x61)) {
				// Track power off
				setPower(csTrackVoltageOff);
			  }
			  else if ((XNetMsg[XNetdata1] == 0x08)) {
				  // Track Short
				  setPower(csShortCircuit);
			  }
			  else if ((XNetMsg[XNetdata1] == 0x02) && (XNetMsg[XNetdata2] == 0x63)) {
				// Service Mode Entry
				setPower(csServiceMode);
			  }
			} 
			else if (XNetMsg[XNetheader] == 0x81) {
				if ((XNetMsg[XNetdata1] == 0x00) && (XNetMsg[XNetdata2] == 0x81)) {
					//Emergency Stop
					setPower(csEmergencyStop);
				}
			}
			else if (XNetMsg[XNetheader] == 0x05 && XNetMsg[XNetdata1] == 0xF1) {
			  //DCC FAST CLOCK set request
			  /* 0x05 0xF1 TCODE1 TCODE2 TCODE3 TCODE4 [XOR]
				00mmmmmm	TCODE1, mmmmmm = denotes minutes, range 0...59.
				100HHHHH	TCODE2, HHHHH = denotes hours, range 0...23.
				01000www	TCODE3, www = denotes day of week, 0=monday, 1=tuesday, a.s.o.
				110fffff	TCODE4, fffff = denotes speed ratio, range 0..31. (0=stopped)
				*/
			}
		}	//Broadcast END
		else if (XNetMsgCallByte == MY_ADDRESS) {
			switch (XNetMsg[XNetheader]) {	
				case 0x52:	// Some other device asked for an accessory change
						break;
				case 0x61:  //Zustand
						break;
				case 0x62:	//Version
						if (XNetMsg[XNetdata1] == 0x22) {
							switch (XNetMsg[XNetdata2]) {
								case 0x00:	//csNormal
											Railpower = csNormal;
											if (notifyXNetPower)
												notifyXNetPower(Railpower);
											break;
								case 0x01:	//csTrackVoltageOff
											Railpower = csTrackVoltageOff;
											if (notifyXNetPower)
												notifyXNetPower(Railpower);
											break;
								case 0x02:	//csEmergencyStop
											Railpower = csEmergencyStop;
											if (notifyXNetPower)
												notifyXNetPower(Railpower);
											break;
								case 0x08:	//csServiceMode
											Railpower = csServiceMode;
											if (notifyXNetPower)
												notifyXNetPower(Railpower);
											break;
							}
							uint8_t commandVersionSequence[] = { 0x00, 0x21, 0x21, 0x00};
							XNetsend (commandVersionSequence, 4);
						}
						break;
				case 0x63:	
						break;
				case 0xE3:	//Antwort abgefrage Funktionen F13-F28
						break;
				case 0xE4:	//Antwort der abgefragen Lok
						break;
				case 0x42:	//Antwort Schaltinformation
						break;
				case 0xE1:  //Err Lok control
						break;
			}	//switch HEADER END
		}	//MY_ADDRESS END
	}	//ENDE SLAVE MODE
}

//--------------------------------------------------------------------------------------------
//Befehl in Zentrale nicht vorhanden
void XpressNetMasterClass::unknown(void)		//unbekannte Anfrage
{
	if (XNetSlaveMode == 0x00) {		//MASTER MODE
		uint8_t NotIn[] = {DirectedOps, 0x61, 0x82, 0xE3 };	
		XNetsend(NotIn, 4);
	}
}

//--------------------------------------------------------------------------------------------
void XpressNetMasterClass::getNextXNetAdr(void)
{
	XNetAdr++;		//nchste Adresse im XNet
	if (XNetAdr > 31) {	//wenn letzte erreicht von Beginn!
		//search for used slots to call them more then unused:
		while((SlotLokUse[XNetAdr % 32] == 0xFFFF) && (XNetAdr > 31)) {	//slot used an loco address?
			XNetAdr++;
		}
		if ((XNetAdr % 32) == 0)
			XNetAdr = 1;	//start at the beginning

	}
	/*
	AAAAA = XNetAdr % 32 (Slot 1 to 31)
	Call Byte Definitions: 
	Call_Response 					P11A AAAA
	Call_Inquiry					P10A AAAA
	Call_Request_Ack_From_Device	P00A AAAA
	Call_Broadcast					P110 0000
	Call_Transfer_Err				P11A AAAA
	*/
	//Generate byte for the new window:
	CallByteInquiry = callByteParity((XNetAdr % 32) | 0x40); // | 0x100;	// the address for Call Byte Window
	RequestAck = callByteParity((XNetAdr % 32)| 0x00); // | 0x100;		// the address for a request acknowlegement sent
	DirectedOps = callByteParity((XNetAdr % 32)| 0x60); // | 0x100;		// the address when we are sending ops
	
	//Send CallByteInquiry for next Addr:
	uint8_t NormalInquiry[] = { CallByteInquiry };
	XNetsend(NormalInquiry, 1);
}

//--------------------------------------------------------------------------------------------
//Zustand der Gleisversorgung setzten
void XpressNetMasterClass::setPower(byte Power) 
{
	if (Railpower != Power) {
	  Railpower = Power;	//Save Power State
	  switch (Power) {	
	  case csNormal: {
			if (XNetSlaveMode == 0x00) {		//MASTER MODE
				uint8_t NormalPower[] = { GENERAL_BROADCAST, 0x61, 0x01, 0x60 };
				XNetsend(NormalPower, 4);
			}
			else {
				uint8_t PowerAn[] = {0x00, 0x21, 0x81, 0xA0 };
				XNetsend(PowerAn, 4);
			}
			break;
		}
	  case csEmergencyStop: {
			if (XNetSlaveMode == 0x00) {		//MASTER MODE
				uint8_t EStopPower[] = { GENERAL_BROADCAST, 0x81, 0x00, 0x81 };
				XNetsend(EStopPower, 4);
			}
			else {
				uint8_t EmStop[] = {0x00, 0x80, 0x80 };
				XNetsend(EmStop, 3);
			}
			break;
		}
	  case csTrackVoltageOff: {
			if (XNetSlaveMode == 0x00) {		//MASTER MODE
				uint8_t OffPower[] = { GENERAL_BROADCAST, 0x61, 0x00, 0x61 };
				XNetsend(OffPower, 4);
			}
			else {
				uint8_t PowerAus[] = {0x00, 0x21, 0x80, 0xA1 };
				XNetsend(PowerAus, 4);
			}
			break;
		}
	  case csShortCircuit: {
			uint8_t ShortPower[] = { GENERAL_BROADCAST, 0x61, 0x08, 0x69 };
			XNetsend(ShortPower, 4);
			break;
		}
	  case csServiceMode: {
			uint8_t ServicePower[] = { GENERAL_BROADCAST, 0x61, 0x02, 0x63 };
			XNetsend(ServicePower, 4);
			break;
		}
	  }
	  if (notifyXNetPower)
			notifyXNetPower(Railpower);
	}
}

//--------------------------------------------------------------------------------------------
//Lok in use (Request/Vormerken)
void XpressNetMasterClass::ReqLocoBusy(uint16_t Adr) {
	//SetLocoBusyAdr = Adr;
	//BusyAdrCount = 0;
	AddBusySlot(0, Adr);
}

//--------------------------------------------------------------------------------------------
//Lok in use (Busy)
void XpressNetMasterClass::SetLocoBusy(uint8_t UserOps, uint16_t Adr) {
	uint8_t LocoInfo[] = { UserOps, 0xE3, 0x40, 0x00, 0x00, 0x00 };
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 6);
	XNetsend(LocoInfo, 6);
}

//--------------------------------------------------------------------------------------------
//Lokinfo an XNet Melden
void XpressNetMasterClass::SetLocoInfo(uint8_t UserOps, uint8_t Speed, uint8_t F0, uint8_t F1) {
	SetLocoInfo(UserOps, Fahrstufe, Speed, F0, F1);
}

void XpressNetMasterClass::SetLocoInfo(uint8_t UserOps, uint8_t Steps, uint8_t Speed, uint8_t F0, uint8_t F1) {
	//0xE4	| 0000 BFFF | RVVV VVVV | 000F FFFF | FFFF FFFF | XOr
	//LokInfo Identification Speed FA FB
	uint8_t LocoInfo[] = {UserOps, 0xE4, Steps, Speed, F0, F1, 0x00 };	
	if (SlotLokUse[UserOps & 0x1F] == 0x00)	//has Slot a Address? 
		LocoInfo[2] |= 0x08;	//set BUSY
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//LokFkt an XNet Melden
void XpressNetMasterClass::SetFktStatus(uint8_t UserOps, uint8_t F4, uint8_t F5) {
	//F4 = F20-F13
	//F5 = F28-F21
	uint8_t LocoFkt[] = {UserOps, 0xE3, 0x52, F4, F5, 0x00 };
	getXOR(LocoFkt, 6);
	XNetsend(LocoFkt, 6);
}

//--------------------------------------------------------------------------------------------
//Lokinfo for MultiMaus
void XpressNetMasterClass::SetLocoInfoMM(uint8_t UserOps, uint8_t Steps, uint8_t Speed, uint8_t F0, uint8_t F1, uint8_t F2, uint8_t F3) {
	//E7 0x04=128 Steps||0x0C=128+Busy (Dir,Speed) (F0,F4,F3,F2,F1) F5-F12 F13-F20 0x00 0x00 CRC
	uint8_t LocoInfo[] = {UserOps,  0xE7, Steps, Speed, 0x20, F1, F2, F3, 0x00, 0x00 };
	LocoInfo[4] |= F0;
	getXOR(LocoInfo, 10);
	XNetsend(LocoInfo, 10);
}

//--------------------------------------------------------------------------------------------
void XpressNetMasterClass::setSpeed(uint16_t Adr, uint8_t Steps, uint8_t Speed) {
	uint8_t LocoInfo[] = {0x00, 0xE4, 0x13, 0x00, 0x00, Speed, 0x00 }; //default to 128 Steps!
	switch (Steps) {
		case 14: LocoInfo[2] = 0x10; break;
		case 27: LocoInfo[2] = 0x11; break;
		case 28: LocoInfo[2] = 0x12; break;
	}
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//Gruppe 1: 0 0 0 F0 F4 F3 F2 F1
void XpressNetMasterClass::setFunc0to4(uint16_t Adr, uint8_t G1) { 
	uint8_t LocoInfo[] = {0x00, 0xE4, 0x20, 0x00, 0x00, G1, 0x00 };
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//Gruppe 2: 0 0 0 0 F8 F7 F6 F5 
void XpressNetMasterClass::setFunc5to8(uint16_t Adr, uint8_t G2) { 
	uint8_t LocoInfo[] = {0x00, 0xE4, 0x21, 0x00, 0x00, G2, 0x00 };
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//Gruppe 3: 0 0 0 0 F12 F11 F10 F9 
void XpressNetMasterClass::setFunc9to12(uint16_t Adr, uint8_t G3) { 
	uint8_t LocoInfo[] = {0x00, 0xE4, 0x22, 0x00, 0x00, G3, 0x00 };
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//Gruppe 4: F20 F19 F18 F17 F16 F15 F14 F13  
void XpressNetMasterClass::setFunc13to20(uint16_t Adr, uint8_t G4) { 
	uint8_t LocoInfoMM[] = {0x00, 0xE4, 0xF3, 0x00, 0x00, G4, 0x00 };	//normal: 0x23!
	LocoInfoMM[3] = Adr >> 8;
	LocoInfoMM[4] = Adr & 0xFF;
	getXOR(LocoInfoMM, 7);
	XNetsend(LocoInfoMM, 7);

	LocoInfoMM[0] = 0x00;
	LocoInfoMM[1] = 0xE4;
	LocoInfoMM[2] = 0x23; 	//MultiMaus only
	LocoInfoMM[3] = Adr >> 8;
	LocoInfoMM[4] = Adr & 0xFF;
	LocoInfoMM[5] = G4;
	getXOR(LocoInfoMM, 7);
	XNetsend(LocoInfoMM, 7);
}

//--------------------------------------------------------------------------------------------
//Gruppe 5: F28 F27 F26 F25 F24 F23 F22 F21  
void XpressNetMasterClass::setFunc21to28(uint16_t Adr, uint8_t G5) { 
	uint8_t LocoInfo[] = {0x00, 0xE4, 0x28, 0x00, 0x00, G5, 0x00 };
	LocoInfo[3] = Adr >> 8;
	LocoInfo[4] = Adr & 0xFF;
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//check if there are slots with the same loco, set them to busy
void XpressNetMasterClass::SetBusy(uint8_t slot) {
	for (byte s = 1; s < 32; s++) {
		if ((s != slot) && (SlotLokUse[s] == SlotLokUse[slot])) {	//if in other Slot -> set busy
			SetLocoBusy(callByteParity(s | 0x60), SlotLokUse[s]);
			SlotLokUse[s] = 0;	//clean slot that informed as busy & let it activ
		}
	}
}

//--------------------------------------------------------------------------------------------
//Add loco to slot. Slot 0 is reserved for non XpressNet Device
void XpressNetMasterClass::AddBusySlot(uint8_t UserOps, uint16_t Adr) {
	if (Adr == 0)
		return;
	SlotLokUse[UserOps & 0x1F] = Adr;	//store loco that is used
	SetBusy(UserOps & 0x1F);	//make other busy
}


//--------------------------------------------------------------------------------------------
//TrntPos request return
void XpressNetMasterClass::SetTrntStatus(uint8_t UserOps, uint8_t Address, uint8_t Data) {
	if (XNetSlaveMode == 0x00) {		//MASTER MODE
		// data=ITTN ZZZZ	Z=Weichenausgang1+2 (Aktive/Inaktive);
		//0x42 AAAA AAAA ITTN ZZZZ
		//I = is 1, the switching command requested has not been completed
		//TT = turnout group (0-3)
		//N = N=0 is the lower nibble, N=1 the upper nibble
		//Z3 Z2 Z1 Z0 = (Z1,Z0|Z3,Z2) 01 "left", 10 "right"
		uint8_t TrntInfo[] = {UserOps, 0x42, 0x00, 0x00, 0x00 };
		TrntInfo[2] = Address;
		TrntInfo[3] = Data;
		getXOR(TrntInfo, 5);
		XNetsend(TrntInfo, 5);
	}
}

//--------------------------------------------------------------------------------------------
//Trnt Change position
void XpressNetMasterClass::SetTrntPos(uint16_t Address, uint8_t state, uint8_t active) {
	//2. Byte = 0000 ABBP
	//A = Weichenausgang(Spulenspannung EIN/AUS)
	//BB = Adresse des Dekoderport 1..4
	//P = Ausgang (Gerade = 0 / Abzweigen = 1)
	uint8_t TrntInfo[] = {0x00, 0x52, 0x00, 0x00, 0x00 };
	TrntInfo[2] = Address >> 2;
	TrntInfo[3] = (Address & 0x03) << 1;
	TrntInfo[3] |= (active & 0x01) << 3;
	TrntInfo[3] |= state & 0x01;
	getXOR(TrntInfo, 5);
	XNetsend(TrntInfo, 5);
}


//--------------------------------------------------------------------------------------------
// send along a bunch of bytes to the Command Station
void XpressNetMasterClass::XNetsend(byte *dataString, byte byteCount) {
	#if defined (XNetDEBUG)
	if (byteCount > 1) {
		XNetSerial.print("XTX: ");
	}
	#endif
	
	XNetBuffer[XNetBufferStore].length = byteCount;
	
	for (byte i = 0; i < byteCount; i++) {
		XNetBuffer[XNetBufferStore].data[i] = *dataString;	//add data to Buffer
		dataString++;

		#if defined (XNetDEBUG)
		if (byteCount > 1) {
			if (XNetBuffer[XNetBufferStore].data[i] < 0x10)
				XNetSerial.print(" 0x0");
			else XNetSerial.print(" 0x");
			if (i == 0 && XNetSlaveMode == 0x00)  //MASTER-MODE
				XNetSerial.print("1");
			XNetSerial.print(XNetBuffer[XNetBufferStore].data[i], HEX);
		}
		#endif
	}
	
	XNetBufferStore++;
	if (XNetBufferStore >= XNetBufferSize)
		XNetBufferStore = 0;
		
	#if defined (XNetDEBUG)
	if (byteCount > 1)
		XNetSerial.println();
	#endif
}

//--------------------------------------------------------------------------------------------
// calculate the XOR
void XpressNetMasterClass::getXOR (uint8_t *data, byte length) {
	byte XOR = 0x00;
	data++;		//without CallByte
	for (byte i = 0; i < (length-2); i++) { 
        XOR = XOR ^ *data;
		data++;
    }
	*data = XOR;
}

//--------------------------------------------------------------------------------------------
// calculate the parity bit in the call byte for this guy
byte XpressNetMasterClass::callByteParity (byte me) {
 byte parity = (1==0);
 byte vv;
 me &= 0x7f;
 vv = me;

 while (vv) {
       parity = !parity;
       vv &= (vv-1);
 }
 if (parity) me |= 0x80;
 return me;
}

//--------------------------------------------------------------------------------------------
//Interrupt routine for writing via Serial
#ifdef SERIAL_PORT_0
ISR(USART_TX_vect) {
	XpressNetMasterClass::handle_TX_interrupt();	 //weiterreichen an die Funktion
}

#else
ISR(USART1_TX_vect) {
	XpressNetMasterClass::handle_TX_interrupt();	 //weiterreichen an die Funktion
}
#endif

// Interrupt handling for receive Data
// static 
inline void XpressNetMasterClass::handle_TX_interrupt()
{
  if (active_object)
  {
    active_object->XNetSendNext();	//nchste Byte Senden
  }
}


//--------------------------------------------------------------------------------------------
void XpressNetMasterClass::XNetSendNext(void) {
	
	uint16_t data9 = XNetReadBuffer();
	if (data9 > 0x1FF) {
		//nothing less to send out. 
		digitalWrite(MAX485_CONTROL, LOW); 	//RECEIVE_MODE
		if (XNet_state == XNet_send_data && XNetSlaveMode == 0x00)
			XNet_state = XNet_get_callbyte;
		else XNet_state = XNet_wait_receive;	//wait for receive from client
		return;
	}
	digitalWrite(MAX485_CONTROL, HIGH); 	//SEND_MODE
		
#ifdef SERIAL_PORT_0
		/* wait for empty transmit buffer */
		//while (!(UCSR0A & (1 << UDRE0))) {}
		/* put the data into buffer, and send */
		UCSRB &= ~(1 << TXB8);
		if (data9 & 0x100) //is there a 9th bit?
			UCSRB |= (1 << TXB8);
		UDR = data9;
#else
		/* wait for empty transmit buffer */
		//while (!(UCSR1A & (1 << UDRE1))) {}
		/* put the data into buffer, and send */
		UCSR1B &= ~(1 << TXB81);
		if (data9 & 0x100)	//is there a 9th bit?
			UCSR1B |= (1 << TXB81);
		UDR1 = data9;
#endif

}

//--------------------------------------------------------------------------------------------
uint16_t XpressNetMasterClass::XNetReadBuffer() {
	if (XNetBuffer[XNetBufferSend].length == 0x00)
		return 0xFFFF;	//no data in Buffer!
	
	uint16_t data = XNetBuffer[XNetBufferSend].data[XNetBufferSend_data];
	if (XNetBufferSend_data == 0x00) {	//it is a CALLBYTE and we are MASTER!
		data |= 0x100;	//add 9th bit

		if (XNetSlaveMode != 0x00) { //SLAVE-MODE?
			XNetBufferSend_data++;	//next data byte
			return 0XFFFF;	//stop here, just send one packet as answer!
		}
	}

	XNetBufferSend_data++;	//next data byte
	if (XNetBufferSend_data >= XNetBuffer[XNetBufferSend].length) {	//any byte left to send?
		XNetBufferSend_data = 0;	//Reset data counter
		XNetBuffer[XNetBufferSend].length = 0x00; //Reset Bufferstore
		XNetBufferSend++;	//next byte
		if (XNetBufferSend >= XNetBufferSize)	//overflow?
			XNetBufferSend = 0;	//start value
	}
	return data;
}

//--------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------
//Interrupt routine for reading via Serial
#ifdef SERIAL_PORT_0
ISR(USART_RX_vect)  {
	XpressNetMasterClass::handle_RX_interrupt();	 //weiterreichen an die Funktion
}

#else
ISR(USART1_RX_vect) {
	XpressNetMasterClass::handle_RX_interrupt();	 //weiterreichen an die Funktion
}
#endif

// Interrupt handling for receive Data
// static 
inline void XpressNetMasterClass::handle_RX_interrupt()
{
  if (active_object)
  {
	active_object->XNetReceive();	//Byte lesen
  }
}

//--------------------------------------------------------------------------------------------
//Serial einlesen:
void XpressNetMasterClass::XNetReceive(void)
{
	
	/*
	unsigned char status;
	#ifdef SERIAL_PORT_0
	status = UCSRA;
	//while (!(status & (1 << RXC))) { return; }
	if (status & ((1 << FE) | (1 << DOR) | (1 << PE))) { XNetSerial.println(UDR, HEX); return; }
	#else
	status = UCSR1A;
	//while (!(status & (1 << RXC1))) { XNetSerial.println(UDR1, HEX); return; }
	if (status & ((1 << FE1) | (1 << DOR1) | (1 << PE1))) { XNetSerial.println(UDR1, HEX); return; }
	#endif
	*/
	
	// if there's any serial available, read it:
	#ifdef SERIAL_PORT_0
		// Filter the 9th bit, then return 
		if (UCSRB & (1 << RXB8)) {
			XNetMsgCallByte = UDR;	//only store data
			XNetSlaveMode = XNetSlaveCycle;	//reactivate SLAVE MODE
		//	XNetMsg[XNetlength] = 0x00;
			if (XNetMsgCallByte == MY_ADDRESS)
				XNetSendNext();	//start sending out by interrupt
		}
		else {
			XNetMsg[XNetlength]++;	    //weitere Nachrichtendaten
			XNetMsg[XNetMsg[XNetlength]] = UDR;                   // Zeichen aus UDR an Aufrufer zurueckgeben
			rxXOR = rxXOR ^ XNetMsg[XNetMsg[XNetlength]];		//XOR bestimmen
		}
	#else
		// Filter the 9th bit, then return 
		if (UCSR1B & (1 << RXB81)) {
			XNetMsgCallByte = UDR1;	//only store data
			XNetSlaveMode = XNetSlaveCycle;	//reactivate SLAVE MODE
		//	XNetMsg[XNetlength] = 0x00;
			if (XNetMsgCallByte == MY_ADDRESS)
				XNetSendNext();	//start sending out by interrupt
		}
		else {
			XNetMsg[XNetlength]++;	    //weitere Nachrichtendaten
			XNetMsg[XNetMsg[XNetlength]] = UDR1;                   // Zeichen aus UDR an Aufrufer zurueckgeben
			rxXOR = rxXOR ^ XNetMsg[XNetMsg[XNetlength]];		//XOR bestimmen
		}
	#endif
	
	if (XNetMsg[XNetlength] >= XNetdata1) { // header and one data byte or more received
		//Check length - length is inside header but without header and xor!
		if (((XNetMsg[XNetheader] & 0x0F) + 2) == XNetMsg[XNetlength])	//reach data length?
			XNetDataReady = true;
		else {
			if (XNetMsg[XNetlength] >= (XNetMaxDataLength - 1) ) {	//overflow!!!
				XNetclear();
			}
		}	
	}
}

//--------------------------------------------------------------------------------------------
//Lschen des letzten gesendeten Befehls
void XpressNetMasterClass::XNetclear(void)
{
	XNetMsg[XNetlength] = 0x00;	//Reset Message
	XNetMsg[XNetheader] = 0x00;
	XNetMsg[XNetdata1] = 0x00;
	XNetMsg[XNetdata2] = 0x00;
	XNetMsg[XNetdata3] = 0x00;
	XNetMsg[XNetdata4] = 0x00;
	XNetMsg[XNetdata5] = 0x00;
	XNetMsg[XNetdata6] = 0x00;
	XNetMsg[XNetdata7] = 0x00;
	rxXOR = 0x00;	//Reset XOR
	XNetMsgCallByte = 0x00;	//Reset CallByte
}
